﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;
using Newtonsoft.Json;
using Pixiview.Resources;
using Pixiview.UI;
using Pixiview.UI.Theme;
using Pixiview.Utils;
using Xamarin.Essentials;
using Xamarin.Forms;

namespace Pixiview.Illust
{
    public abstract class FavoriteIllustCollectionPage : IllustCollectionPage<IllustItem[]> { }
    public abstract class IllustDataCollectionPage : IllustCollectionPage<IllustData> { }
    public abstract class IllustRankingDataCollectionPage : IllustCollectionPage<IllustRankingData> { }
    public abstract class IllustUserDataCollectionPage : IllustCollectionPage<IllustUserData> { }

    public interface IIllustCollectionPage
    {
        IllustCollection IllustCollection { get; set; }
    }
    public abstract class IllustCollectionPage<T> : AdaptedPage, IIllustCollectionPage
    {
        #region - Properties -

        public static readonly BindableProperty IllustsProperty = BindableProperty.Create(
            nameof(Illusts), typeof(IllustCollection), typeof(IllustCollectionPage<T>));
        public static readonly BindableProperty ColumnsProperty = BindableProperty.Create(
            nameof(Columns), typeof(int), typeof(IllustCollectionPage<T>), 2);
        public static readonly BindableProperty IsLoadingProperty = BindableProperty.Create(
            nameof(IsLoading), typeof(bool), typeof(IllustCollectionPage<T>), true);

        public IllustCollection Illusts
        {
            get => (IllustCollection)GetValue(IllustsProperty);
            set => SetValue(IllustsProperty, value);
        }
        public int Columns
        {
            get => (int)GetValue(ColumnsProperty);
            set => SetValue(ColumnsProperty, value);
        }
        public bool IsLoading
        {
            get => (bool)GetValue(IsLoadingProperty);
            set => SetValue(IsLoadingProperty, value);
        }

        #endregion

        protected static DateTime LastUpdated { get; private set; } = DateTime.Now;

        public IllustCollection IllustCollection { get; set; }

        protected virtual bool IsFavoriteVisible => true;
        protected readonly Command<IllustItem> commandIllustImageTapped;
        protected DateTime lastUpdated;
        protected double topOffset;

        private T illustData;
        private readonly ParallelOptions parallelOptions = new ParallelOptions { MaxDegreeOfParallelism = Configs.MaxThreads };

        public IllustCollectionPage()
        {
            commandIllustImageTapped = new Command<IllustItem>(IllustImageTapped);
            BindingContext = this;
        }

        private void IllustImageTapped(IllustItem illust)
        {
            if (illust == null)
            {
                return;
            }
            Start(() => OnIllustImageTapped(illust));
        }

        #region - Overrides -

        public override void OnUnload()
        {
            Illusts = IllustCollection.Empty;
            lastUpdated = default;
        }

        protected override void OnAppearing()
        {
            base.OnAppearing();
            Connectivity.ConnectivityChanged += Connectivity_ConnectivityChanged;

            if (lastUpdated != LastUpdated)
            {
                Device.StartTimer(TimeSpan.FromMilliseconds(200), () =>
                {
                    StartLoad();
                    return false;
                });
            }
            else if (IsFavoriteVisible && IllustCollection != null)
            {
                var favorites = Stores.Favorites;
                foreach (var item in IllustCollection)
                {
                    item.IsFavorite = favorites.Any(i => i.Id == item.Id);
                }
            }
        }

        protected override void OnDisappearing()
        {
            Connectivity.ConnectivityChanged -= Connectivity_ConnectivityChanged;
            base.OnDisappearing();
        }

        private void Connectivity_ConnectivityChanged(object sender, ConnectivityChangedEventArgs e)
        {
            if (e.NetworkAccess == NetworkAccess.Internet)
            {
                if (lastUpdated != LastUpdated)
                {
                    StartLoad(true);
                }
            }
        }

        protected override void OnSizeAllocated(double width, double height)
        {
            base.OnSizeAllocated(width, height);
            int columns;
#if __IOS__
            var oldMargin = PanelTopMargin;
            Thickness newMargin;
            if (StyleDefinition.IsFullscreenDevice)
            {
                newMargin = width > height ?
                    AppShell.HalfNavigationBarOffset :
                    AppShell.NavigationBarOffset;
                topOffset = width > height ?
                    AppShell.NavigationBarOffset.Top :
                    AppShell.TotalBarOffset.Top;
            }
            else if (isPhone)
            {
                newMargin = width > height ?
                    StyleDefinition.TopOffset16 :
                    StyleDefinition.TopOffset32;
                topOffset = width > height ?
                    StyleDefinition.TopOffset32.Top :
                    AppShell.TotalBarOffset.Top;
            }
            else
            {
                // ipad
                newMargin = StyleDefinition.TopOffset37;
                topOffset = AppShell.TotalBarOffset.Top;
            }
#endif
            if (width > height)
            {
                columns = isPhone ? 4 : 6;
            }
            else
            {
                columns = isPhone ? 2 : 4;
            }
#if __IOS__
            if (oldMargin != newMargin)
            {
                PanelTopMargin = newMargin;
                OnPanelTopMarginChanged(oldMargin, newMargin);
            }
#endif
            if (Columns != columns)
            {
                Columns = columns;
                App.DebugPrint($"change columns to {columns}");
            }
        }

        #endregion

        protected abstract T DoLoadIllustData(bool force);
        protected abstract IEnumerable<IllustItem> DoGetIllustList(T data);
        protected virtual void OnIllustImageTapped(IllustItem illust)
        {
            var page = new ViewIllustPage(illust, IsFavoriteVisible);
            Navigation.PushAsync(page);
        }
        protected virtual void DoIllustsLoaded(IllustCollection collection)
        {
            IllustCollection = collection;
            Illusts = collection;
        }

        protected void StartLoad(bool force = false)
        {
            if (force || lastUpdated != LastUpdated)
            {
                lastUpdated = LastUpdated;
                IsLoading = true;
                Task.Run(() => DoLoadIllusts(force));
            }
        }

        protected DataTemplate GetCardViewTemplate(bool hideUser = false)
        {
            return new DataTemplate(() =>
            {
                #region - components -

                // image
                var image = new RoundImage
                {
                    BackgroundColor = StyleDefinition.ColorDownloadBackground,
                    CornerRadius = 10,
                    CornerMasks = CornerMask.Top,
                    HorizontalOptions = LayoutOptions.Fill,
                    Aspect = Aspect.AspectFill,
                    GestureRecognizers =
                    {
                        new TapGestureRecognizer
                        {
                            Command = commandIllustImageTapped
                        }
                        .Binding(TapGestureRecognizer.CommandParameterProperty, ".")
                    }
                }
                .Binding(Image.SourceProperty, nameof(IllustItem.Image));

                // label: r-18
                var r18 = new RoundLabel
                {
                    Text = ResourceHelper.R18,
                    BackgroundColor = StyleDefinition.ColorRedBackground,
                    Margin = new Thickness(6, 6, 0, 0),
                    Padding = new Thickness(6, 2),
                    CornerRadius = 4,
                    HorizontalOptions = LayoutOptions.Start,
                    VerticalOptions = LayoutOptions.Start,
                    FontSize = StyleDefinition.FontSizeMicro,
                    TextColor = Color.White
                }
                .Binding(IsVisibleProperty, nameof(IllustItem.IsRestrict));

                // label: pages
                var pages = new RoundLabel
                {
                    BackgroundColor = StyleDefinition.ColorDeepShadow,
                    Margin = new Thickness(0, 6, 6, 0),
                    Padding = new Thickness(6, 4),
                    CornerRadius = 6,
                    HorizontalOptions = LayoutOptions.End,
                    VerticalOptions = LayoutOptions.Start,
                    FontSize = StyleDefinition.FontSizeMicro,
                    TextColor = Color.White
                }
                .Binding(Label.TextProperty, nameof(IllustItem.PageCountText))
                .Binding(IsVisibleProperty, nameof(IllustItem.IsPageVisible))
                .DynamicResource(Label.FontFamilyProperty, ThemeBase.IconSolidFontFamily);

                // label: is anime
                var anime = new RoundLabel
                {
                    Text = StyleDefinition.IconPlay,
                    BackgroundColor = StyleDefinition.ColorDeepShadow,
                    Margin = new Thickness(0, 0, 6, 6),
                    Padding = new Thickness(12, 9, 0, 0),
                    WidthRequest = 36,
                    HeightRequest = 36,
                    CornerRadius = 18,
                    HorizontalOptions = LayoutOptions.End,
                    VerticalOptions = LayoutOptions.End,
                    FontSize = StyleDefinition.FontSizeTitle,
                    TextColor = Color.White
                }
                .Binding(IsVisibleProperty, nameof(IllustItem.IsAnimeVisible))
                .DynamicResource(Label.FontFamilyProperty, ThemeBase.IconSolidFontFamily);

                // label: title
                var title = new Label
                {
                    Padding = new Thickness(8, 2),
                    HorizontalOptions = LayoutOptions.FillAndExpand,
                    VerticalOptions = LayoutOptions.Center,
                    LineBreakMode = LineBreakMode.TailTruncation,
                    FontSize = StyleDefinition.FontSizeSmall
                }.DynamicResource(Label.TextColorProperty, ThemeBase.TextColor);

                // label: favorite
                var favorite = new Label
                {
                    WidthRequest = 26,
                    HorizontalOptions = LayoutOptions.End,
                    HorizontalTextAlignment = TextAlignment.End,
                    VerticalOptions = LayoutOptions.Center,
                    FontSize = StyleDefinition.FontSizeSmall,
                    TextColor = StyleDefinition.ColorRedBackground,
                    Text = StyleDefinition.IconLove
                }
                .Binding(IsVisibleProperty, IsFavoriteVisible ? nameof(IllustItem.IsFavorite) : null)
                .DynamicResource(Label.FontFamilyProperty, ThemeBase.IconSolidFontFamily);

                #endregion

                if (hideUser)
                {
                    return new CardView
                    {
                        Padding = 0,
                        Margin = 0,
                        CornerRadius = 10,
                        ShadowColor = StyleDefinition.ColorLightShadow,
                        ShadowOffset = new Size(2, 2),
                        Content = new Grid
                        {
                            HorizontalOptions = LayoutOptions.Fill,
                            RowSpacing = 0,
                            RowDefinitions =
                            {
                                new RowDefinition().Binding(RowDefinition.HeightProperty, nameof(IllustItem.ImageHeight)),
                                new RowDefinition { Height = 30 }
                            },
                            Children =
                            {
                                image,
                                r18,
                                pages,
                                anime,

                                // stacklayout: title
                                new Grid
                                {
                                    ColumnDefinitions =
                                    {
                                        new ColumnDefinition(),
                                        new ColumnDefinition { Width = 20 }
                                    },
                                    VerticalOptions = LayoutOptions.Center,
                                    Padding = new Thickness(0, 0, 8, 0),
                                    Children =
                                    {
                                        title.Binding(Label.TextProperty, nameof(IllustItem.Title)),
                                        favorite.GridColumn(1)
                                    }
                                }
                                .GridRow(1)
                            }
                        }
                    }
                    .DynamicResource(BackgroundColorProperty, ThemeBase.CardBackgroundColor);
                }

                return new CardView
                {
                    Padding = 0,
                    Margin = 0,
                    CornerRadius = 10,
                    ShadowColor = StyleDefinition.ColorLightShadow,
                    ShadowOffset = new Size(2, 2),
                    Content = new Grid
                    {
                        HorizontalOptions = LayoutOptions.Fill,
                        RowSpacing = 0,
                        RowDefinitions =
                        {
                            new RowDefinition().Binding(RowDefinition.HeightProperty, nameof(IllustItem.ImageHeight)),
                            new RowDefinition { Height = 30 },
                            new RowDefinition { Height = 40 }
                        },
                        Children =
                        {
                            image,
                            r18,
                            pages,
                            anime,

                            title.Binding(Label.TextProperty, nameof(IllustItem.RankTitle)).GridRow(1),

                            // stacklayout: user
                            new Grid
                            {
                                ColumnDefinitions =
                                {
                                    new ColumnDefinition { Width = 30 },
                                    new ColumnDefinition(),
                                    new ColumnDefinition { Width = 20 }
                                },
                                Padding = new Thickness(8, 0, 8, 8),
                                Children =
                                {
                                    // user icon
                                    new CircleImage
                                    {
                                        WidthRequest = 30,
                                        HeightRequest = 30,
                                        Aspect = Aspect.AspectFill
                                    }
                                    .Binding(Image.SourceProperty, nameof(IllustItem.ProfileImage)),

                                    // user name
                                    new Label
                                    {
                                        HorizontalOptions = LayoutOptions.FillAndExpand,
                                        VerticalOptions = LayoutOptions.Center,
                                        LineBreakMode = LineBreakMode.TailTruncation,
                                        FontSize = StyleDefinition.FontSizeMicro
                                    }
                                    .Binding(Label.TextProperty, nameof(IllustItem.UserName))
                                    .DynamicResource(Label.TextColorProperty, ThemeBase.SubTextColor)
                                    .GridColumn(1),

                                    // label: favorite
                                    favorite.GridColumn(2)
                                }
                            }
                            .GridRow(2)
                        }
                    }
                }
                .DynamicResource(BackgroundColorProperty, ThemeBase.CardBackgroundColor);
            });
        }

        #region - Illust Tasks -

        protected void DoLoadIllusts(bool force = false)
        {
            illustData = DoLoadIllustData(force);
            if (illustData == null)
            {
                //App.DebugError("illusts.load", "failed to load illusts data.");
                IsLoading = false;
                return;
            }
            if (force && IsFavoriteVisible)
            {
                var now = DateTime.Now;
                LastUpdated = now;
                lastUpdated = now;
            }

            var data = DoGetIllustList(illustData).Where(i => i != null);

            var collection = new IllustCollection(data);
            if (IsFavoriteVisible)
            {
                var favorites = Stores.Favorites;
                foreach (var item in collection)
                {
                    if (item.Image == null)
                    {
                        item.Image = StyleDefinition.DownloadBackground;
                    }
                    item.IsFavorite = favorites.Any(i => i.Id == item.Id);
                }
            }
            DoIllustsLoaded(collection);
            IsLoading = false;

            DoLoadImages(collection);
        }

        void DoLoadImages(IllustCollection collection)
        {
            Parallel.ForEach(collection, parallelOptions, illust =>
            {
                if (!collection.Running)
                {
                    return;
                }
                if (illust.ImageUrl != null)
                {
                    var url = Configs.GetThumbnailUrl(illust.ImageUrl);
                    var image = Stores.LoadPreviewImage(url, false);
                    if (image == null)
                    {
                        image = Stores.LoadThumbnailImage(url);
                    }
                    if (image != null)
                    {
                        illust.Image = image;
                    }
                }

                if (illust.ProfileUrl != null)
                {
                    var userImage = Stores.LoadUserProfileImage(illust.ProfileUrl);
                    if (userImage != null)
                    {
                        illust.ProfileImage = userImage;
                    }
                }
            });
        }

        #endregion
    }

    public enum ScrollDirection
    {
        Stop,
        Up,
        Down
    }

    public class IllustCollection : List<IllustItem>
    {
        private static IllustCollection empty;

        public static IllustCollection Empty
        {
            get
            {
                if (empty == null)
                {
                    empty = new IllustCollection();
                }
                return empty;
            }
        }

        public IllustCollection() : base()
        {
            running = true;
        }
        public IllustCollection(IEnumerable<IllustItem> illusts) : base(illusts)
        {
            running = true;
        }

        private readonly object sync = new object();
        private volatile bool running;
        public bool Running
        {
            get => running;
            set
            {
                lock (sync)
                {
                    running = value;
                }
            }
        }
    }

    public class IllustFavorite
    {
        public DateTime LastFavoriteUtc { get; set; }
        public List<IllustItem> Illusts { get; set; }
    }

    public enum IllustType
    {
        Illust = 0,
        Manga = 1,
        Anime = 2
    }

    public interface IIllustUser
    {
        string UserId { get; }
        string UserName { get; }
        ImageSource ProfileImage { get; }
    }

    [JsonObject(MemberSerialization.OptIn)]
    public class IllustItem : BindableObject, IIllustUser
    {
        private static IllustItem empty;
        public static IllustItem Empty
        {
            get
            {
                if (empty == null)
                {
                    empty = new IllustItem();
                }
                return empty;
            }
        }

        public static readonly BindableProperty ImageProperty = BindableProperty.Create(
            nameof(Image), typeof(ImageSource), typeof(IllustItem));
        public static readonly BindableProperty ProfileImageProperty = BindableProperty.Create(
            nameof(ProfileImage), typeof(ImageSource), typeof(IllustItem));
        public static readonly BindableProperty ImageHeightProperty = BindableProperty.Create(
            nameof(ImageHeight), typeof(GridLength), typeof(IllustItem), GridLength.Auto);
        public static readonly BindableProperty IsFavoriteProperty = BindableProperty.Create(
            nameof(IsFavorite), typeof(bool), typeof(IllustItem));

        [JsonProperty]
        public string Title { get; set; }
        public string RankTitle { get; set; }
        public ImageSource Image
        {
            get => (ImageSource)GetValue(ImageProperty);
            set => SetValue(ImageProperty, value);
        }
        public ImageSource ProfileImage
        {
            get => (ImageSource)GetValue(ProfileImageProperty);
            set => SetValue(ProfileImageProperty, value);
        }
        public GridLength ImageHeight
        {
            get => (GridLength)GetValue(ImageHeightProperty);
            set => SetValue(ImageHeightProperty, value);
        }
        public bool IsFavorite
        {
            get => (bool)GetValue(IsFavoriteProperty);
            set => SetValue(IsFavoriteProperty, value);
        }
        public bool IsPlaying { get; set; }

        [JsonProperty]
        public string Id { get; set; }
        [JsonProperty]
        public string ImageUrl { get; set; }
        [JsonProperty]
        public bool IsRestrict { get; set; }
        [JsonProperty]
        public string ProfileUrl { get; set; }
        [JsonProperty]
        public string UserId { get; set; }
        [JsonProperty]
        public string UserName { get; set; }
        [JsonProperty]
        public int Width { get; set; }
        [JsonProperty]
        public int Height { get; set; }
        [JsonProperty]
        public int PageCount { get; set; }
        [JsonProperty]
        public double ImageHeightValue
        {
            get => ImageHeight.IsAuto ? -1 : ImageHeight.Value;
            set => ImageHeight = value > 0 ? value : GridLength.Auto;
        }
        [JsonProperty]
        public int YesRank { get; set; }
        [JsonProperty]
        public int RatingCount { get; set; }
        [JsonProperty]
        public int ViewCount { get; set; }
        [JsonProperty]
        public long UploadTimestamp { get; set; }
        [JsonProperty]
        public IllustType IllustType { get; set; }

        public string PageCountText => $"{StyleDefinition.IconLayer} {PageCount}";
        public bool IsPageVisible => PageCount > 1;
        public bool IsAnimeVisible => IllustType == IllustType.Anime;
    }
}
