﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;
using Pixiview.Resources;
using Pixiview.UI;
using Pixiview.UI.Theme;
using Pixiview.Utils;
using Xamarin.Essentials;
using Xamarin.Forms;

namespace Pixiview.Illust
{
    [QueryProperty("IllustId", "id")]
    public partial class ViewIllustPage : AdaptedPage
    {
        public static readonly BindableProperty FavoriteIconProperty = BindableProperty.Create(
            nameof(FavoriteIcon), typeof(ImageSource), typeof(ViewIllustPage));

        public static readonly BindableProperty IllustsProperty = BindableProperty.Create(
            nameof(Illusts), typeof(IllustDetailItem[]), typeof(ViewIllustPage));
        public static readonly BindableProperty PagePositionTextProperty = BindableProperty.Create(
            nameof(PagePositionText), typeof(string), typeof(ViewIllustPage));
        public static readonly BindableProperty CurrentPageProperty = BindableProperty.Create(
            nameof(CurrentPage), typeof(int), typeof(ViewIllustPage), propertyChanged: OnCurrentPagePropertyChanged);
        public static readonly BindableProperty IsScrollAnimatedProperty = BindableProperty.Create(
            nameof(IsScrollAnimated), typeof(bool), typeof(ViewIllustPage), true);
        public static readonly BindableProperty IsAnimateSliderVisibleProperty = BindableProperty.Create(
            nameof(IsAnimateSliderVisible), typeof(bool), typeof(ViewIllustPage));
        public static readonly BindableProperty IsAnimateSliderEnabledProperty = BindableProperty.Create(
            nameof(IsAnimateSliderEnabled), typeof(bool), typeof(ViewIllustPage));
        public static readonly BindableProperty CurrentAnimeFrameProperty = BindableProperty.Create(
            nameof(CurrentAnimeFrame), typeof(double), typeof(ViewIllustPage), propertyChanged: OnCurrentAnimeFramePropertyChanged);
        public static readonly BindableProperty MaximumFrameProperty = BindableProperty.Create(
            nameof(MaximumFrame), typeof(double), typeof(ViewIllustPage), 1.0);

        private static void OnCurrentPagePropertyChanged(BindableObject obj, object old, object @new)
        {
            var page = (ViewIllustPage)obj;
            var index = (int)@new;
            var items = page.Illusts;
            var length = items.Length;
            page.PagePositionText = $"{index + 1}/{length}";

            var item = items[index];
            if (!item.Loading && item.Image == null)
            {
                Task.Run(() => page.DoLoadImage(index));
            }
            if (index < length - 1)
            {
                item = items[index + 1];
                if (!item.Loading && item.Image == null)
                {
                    Task.Run(() => page.DoLoadImage(index + 1));
                }
            }
        }

        private static void OnCurrentAnimeFramePropertyChanged(BindableObject obj, object old, object @new)
        {
            var page = (ViewIllustPage)obj;
            if (page.ugoira != null && page.IsAnimateSliderEnabled)
            {
                var frame = (double)@new;
                page.ugoira.ToggleFrame((int)frame);
            }
        }

        public ImageSource FavoriteIcon
        {
            get => (ImageSource)GetValue(FavoriteIconProperty);
            set => SetValue(FavoriteIconProperty, value);
        }

        public IllustDetailItem[] Illusts
        {
            get => (IllustDetailItem[])GetValue(IllustsProperty);
            set => SetValue(IllustsProperty, value);
        }
        public string PagePositionText
        {
            get => (string)GetValue(PagePositionTextProperty);
            set => SetValue(PagePositionTextProperty, value);
        }
        public int CurrentPage
        {
            get => (int)GetValue(CurrentPageProperty);
            private set => SetValue(CurrentPageProperty, value);
        }
        public bool IsScrollAnimated
        {
            get => (bool)GetValue(IsScrollAnimatedProperty);
            set => SetValue(IsScrollAnimatedProperty, value);
        }
        public bool IsAnimateSliderVisible
        {
            get => (bool)GetValue(IsAnimateSliderVisibleProperty);
            set => SetValue(IsAnimateSliderVisibleProperty, value);
        }
        public bool IsAnimateSliderEnabled
        {
            get => (bool)GetValue(IsAnimateSliderEnabledProperty);
            set => SetValue(IsAnimateSliderEnabledProperty, value);
        }
        public double CurrentAnimeFrame
        {
            get => (double)GetValue(CurrentAnimeFrameProperty);
            set => SetValue(CurrentAnimeFrameProperty, value);
        }
        public double MaximumFrame
        {
            get => (double)GetValue(MaximumFrameProperty);
            set => SetValue(MaximumFrameProperty, value);
        }

        public IllustItem IllustItem { get; private set; }
        public bool IsPageVisible { get; private set; }

        private readonly bool saveFavorites;
        private readonly ImageSource fontIconLove;
        private readonly ImageSource fontIconNotLove;
        private IllustUgoiraData ugoiraData;
        private Ugoira ugoira;

        public ViewIllustPage(IllustItem illust, bool save)
        {
            IllustItem = illust;
            saveFavorites = save;
            BindingContext = this;

            fontIconLove = (ImageSource)Application.Current.Resources[ThemeBase.FontIconLove];
            fontIconNotLove = (ImageSource)Application.Current.Resources[ThemeBase.FontIconNotLove];
            var favorites = Stores.Favorites;
            FavoriteIcon = favorites.Any(i => i.Id == illust.Id)
                ? fontIconLove
                : fontIconNotLove;

            var pageVisible = illust != null && illust.PageCount > 1;
            IsPageVisible = pageVisible;
            Resources.Add("carouselView", GetCarouseTemplate(pageVisible));

            InitializeComponent();

            if (illust != null)
            {
                LoadIllust(illust);
            }
        }

        public override void OnLoad()
        {
            OnOrientationChanged(CurrentOrientation);
        }

        protected override void OnAppearing()
        {
            base.OnAppearing();
            Screen.SetHomeIndicatorAutoHidden(Shell.Current, true);
        }

        protected override void OnDisappearing()
        {
            base.OnDisappearing();

            if (ugoira != null)
            {
                ugoira.TogglePlay(false);
                IllustItem.IsPlaying = false;
            }

            if (saveFavorites)
            {
                Stores.SaveFavoritesIllusts();
            }
            Screen.SetHomeIndicatorAutoHidden(Shell.Current, false);
        }

        private DataTemplate GetCarouseTemplate(bool multiPages)
        {
            var tap = new TapGestureRecognizer();
            tap.Tapped += Image_Tapped;

            return new DataTemplate(() =>
            {
                // image
                var image = new Image
                {
                    HorizontalOptions = LayoutOptions.Fill,
                    VerticalOptions = LayoutOptions.Fill,
                    Aspect = Aspect.AspectFit,
                    GestureRecognizers = { tap }
                }
                .Binding(Image.SourceProperty, nameof(IllustDetailItem.Image));

                // downloading
                var downloading = new Frame
                {
                    HasShadow = false,
                    Margin = default,
                    Padding = new Thickness(20),
                    CornerRadius = 8,
                    HorizontalOptions = LayoutOptions.Center,
                    VerticalOptions = LayoutOptions.Center,
                    Content = new ActivityIndicator
                    {
                        IsRunning = true,
                        IsVisible = true
                    }
                    .DynamicResource(ActivityIndicator.ColorProperty, ThemeBase.WindowColor)
                }
                .Binding(IsVisibleProperty, nameof(IllustDetailItem.Loading))
                .DynamicResource(BackgroundColorProperty, ThemeBase.MaskColor);

                // loading original
                var original = new ActivityIndicator
                {
                    IsRunning = true,
                    Margin = new Thickness(10),
                    HorizontalOptions = LayoutOptions.Start,
                    VerticalOptions = LayoutOptions.Start
                }
                .Binding(IsVisibleProperty, nameof(IllustDetailItem.Downloading))
                .DynamicResource(ActivityIndicator.ColorProperty, ThemeBase.TextColor);

                if (multiPages)
                {
                    var tapPrevious = new TapGestureRecognizer();
                    tapPrevious.Tapped += TapPrevious_Tapped;
                    var tapNext = new TapGestureRecognizer();
                    tapNext.Tapped += TapNext_Tapped;

                    return new Grid
                    {
                        Children =
                        {
                            // image
                            image,

                            // tap holder
                            new Grid
                            {
                                RowDefinitions =
                                {
                                    new RowDefinition(),
                                    new RowDefinition(),
                                    new RowDefinition()
                                },
                                Children =
                                {
                                    new Label
                                    {
                                        GestureRecognizers = { tapPrevious }
                                    },
                                    new Label
                                    {
                                        GestureRecognizers = { tapNext }
                                    }
                                    .GridRow(2)
                                }
                            },

                            // downloading
                            downloading,

                            // loading original
                            original
                        }
                    };
                }

                return new Grid
                {
                    Children =
                    {
                        // image
                        image,

                        // downloading
                        downloading,

                        // loading original
                        original
                    }
                };
            });
        }

        private void LoadIllust(IllustItem illust)
        {
            if (illust == null)
            {
                return;
            }

            var items = new IllustDetailItem[illust.PageCount];
            if (items.Length > 1)
            {
                PagePositionText = $"1/{items.Length}";
            }

            for (var i = 0; i < items.Length; i++)
            {
                items[i] = new IllustDetailItem
                {
                    Id = illust.Id
                };
                if (i == 0)
                {
                    items[i].Loading = true;
                    items[i].Image = illust.Image;
                }
            }

            Illusts = items;
            Task.Run(DoLoadImages);
        }

        private void DoLoadImages()
        {
            var illustItem = IllustItem;
            var pages = Stores.LoadIllustPageData(illustItem.Id);
            if (pages == null)
            {
                App.DebugError("illustPage.load", $"failed to load illust page data, id: {illustItem.Id}");
                return;
            }
            var items = Illusts;
            if (pages.body.Length > items.Length)
            {
                App.DebugPrint($"local page count ({items.Length}) is not equals the remote one ({pages.body.Length})");
                var tmp = new IllustDetailItem[pages.body.Length];
                items.CopyTo(items, 0);
                for (var i = items.Length; i < tmp.Length; i++)
                {
                    tmp[i] = new IllustDetailItem
                    {
                        Id = illustItem.Id
                    };
                }
                Illusts = items = tmp;
            }

            for (var i = 0; i < items.Length; i++)
            {
                var item = items[i];
                var p = pages.body[i];
                item.PreviewUrl = p.urls.regular;
                item.OriginalUrl = p.urls.original;

                if (i == 0 && illustItem.ImageUrl == null)
                {
                    // maybe open from a link
                    var preload = Stores.LoadIllustPreloadData(illustItem.Id);
                    if (preload != null && preload.illust.TryGetValue(illustItem.Id, out var illust))
                    {
                        illust.CopyToItem(illustItem);
                        if (preload.user.TryGetValue(illust.userId, out var user))
                        {
                            illustItem.ProfileUrl = user.image;
                        }
                    }
                }
            }

            DoLoadImage(0, true);
            var image = items[0].Image;
            if (image != null)
            {
                illustItem.Image = image;
            }
            if (items.Length > 1)
            {
                DoLoadImage(1);
            }
            else if (illustItem.IllustType == IllustType.Anime)
            {
                // anime
                ugoiraData = Stores.LoadIllustUgoiraData(illustItem.Id);
                if (ugoiraData != null && ugoiraData.body != null)
                {
                    var length = ugoiraData.body.frames.Length;
                    MaximumFrame = length > 0 ? length : 1;
                    IsAnimateSliderVisible = true;
                }
            }
        }

        private void DoLoadImage(int index, bool force = false)
        {
            var items = Illusts;
            if (index < 0 || index >= items.Length)
            {
                App.DebugPrint($"invalid index: {index}");
                return;
            }

            var item = items[index];
            if (!force)
            {
                if (item.Loading || (index > 0 && item.Image != null))
                {
                    App.DebugPrint($"skipped, loading or already loaded, index: {index}, loading: {item.Loading}");
                    return;
                }
            }
            item.Loading = true;
            var image = Stores.LoadPreviewImage(item.PreviewUrl);
            if (image != null)
            {
                item.Image = image;
            }
            item.Loading = false;
        }

        private void CarouselView_PositionChanged(object sender, PositionChangedEventArgs e)
        {
            CurrentPage = e.CurrentPosition;
        }

        private void TapPrevious_Tapped(object sender, EventArgs e)
        {
            var index = CurrentPage;
            if (index > 0)
            {
                IsScrollAnimated = false;
                CurrentPage = index - 1;
                IsScrollAnimated = true;
            }
        }

        private void TapNext_Tapped(object sender, EventArgs e)
        {
            var index = CurrentPage;
            var illusts = Illusts;
            if (illusts != null && index < illusts.Length - 1)
            {
                IsScrollAnimated = false;
                CurrentPage = index + 1;
                IsScrollAnimated = true;
            }
        }

        private void Favorite_Clicked(object sender, EventArgs e)
        {
            var favorites = Stores.Favorites;
            var illust = IllustItem;
            var index = favorites.FindIndex(i => i.Id == illust.Id);
            if (index < 0)
            {
                illust.IsFavorite = true;
                favorites.Insert(0, illust);
                FavoriteIcon = fontIconLove;
            }
            else
            {
                illust.IsFavorite = false;
                favorites.RemoveAt(index);
                FavoriteIcon = fontIconNotLove;
            }
        }

        private void Image_Tapped(object sender, EventArgs e)
        {
            if (ugoiraData == null)
            {
                return;
            }
            var illustItem = IllustItem;

            if (ugoira != null)
            {
                var playing = !ugoira.IsPlaying;
                IsAnimateSliderEnabled = !playing;
                ugoira.TogglePlay(playing);
                illustItem.IsPlaying = playing;
            }
            else if (((Image)sender).BindingContext is IllustDetailItem item)
            {
                if (illustItem.IsPlaying ||
                    illustItem.IllustType != IllustType.Anime)
                {
                    return;
                }

                ugoira = new Ugoira(ugoiraData, item);
                ugoira.FrameChanged += OnUgoiraFrameChanged;
                illustItem.IsPlaying = true;
                ugoira.TogglePlay(true);
            }
        }

        private void OnUgoiraFrameChanged(object sender, UgoiraEventArgs e)
        {
            e.DetailItem.Image = e.Image;
            CurrentAnimeFrame = e.FrameIndex;
        }

        private async void More_Clicked(object sender, EventArgs e)
        {
            int p = CurrentPage;
            var illusts = Illusts;
            if (illusts == null || p < 0 || p >= illusts.Length)
            {
                return;
            }

            var item = illusts[p];

            List<string> extras = new List<string>();
            var share = ResourceHelper.Share;
            var preview = Stores.GetPreviewImagePath(item.PreviewUrl);
            if (preview != null)
            {
                extras.Add(share);
            }
            var saveOriginal = ResourceHelper.SaveOriginal;
            var userDetail = ResourceHelper.UserDetail;
            extras.Add(userDetail);

            var illustItem = IllustItem;
            var result = await DisplayActionSheet(
                illustItem.Title,
                ResourceHelper.Cancel,
                saveOriginal,
                extras.ToArray());
            if (result == saveOriginal)
            {
                SaveOriginalImage(item);
            }
            else if (result == share)
            {
                await Share.RequestAsync(new ShareFileRequest
                {
                    Title = illustItem.Title,
                    File = new ShareFile(preview)
                });
            }
            else if (result == userDetail)
            {
                var page = new UserIllustPage(illustItem);
                await Navigation.PushAsync(page);
            }
        }

        private async void SaveOriginalImage(IllustDetailItem item)
        {
            if (Stores.CheckIllustImage(item.OriginalUrl))
            {
                var flag = await DisplayAlert(ResourceHelper.Operation,
                    ResourceHelper.AlreadySavedQuestion,
                    ResourceHelper.Yes,
                    ResourceHelper.No);
                if (!flag)
                {
                    return;
                }
            }

            var status = await Permissions.CheckStatusAsync<Permissions.Photos>();
            if (status != PermissionStatus.Granted)
            {
                status = await Permissions.RequestAsync<Permissions.Photos>();
                if (status != PermissionStatus.Granted)
                {
                    App.DebugPrint("access denied to gallery.");
                    return;
                }
            }

            if (item == null || item.Downloading)
            {
                return;
            }
            item.Downloading = true;
            _ = Task.Run(() => DoLoadOriginalImage(item));
        }

        private void DoLoadOriginalImage(IllustDetailItem item)
        {
            var image = Stores.LoadIllustImage(item.OriginalUrl);
            if (image != null)
            {
                MainThread.BeginInvokeOnMainThread(async () =>
                {
                    var result = await FileStore.SaveImageToGalleryAsync(image);

                    string message = result ?? ResourceHelper.SaveSuccess;
                    await DisplayAlert(ResourceHelper.Title, message, ResourceHelper.Ok);
                });
            }
            item.Downloading = false;
        }
    }

    public class IllustDetailItem : BindableObject
    {
        public static readonly BindableProperty ImageProperty = BindableProperty.Create(
            nameof(Image), typeof(ImageSource), typeof(IllustDetailItem));
        public static readonly BindableProperty LoadingProperty = BindableProperty.Create(
            nameof(Loading), typeof(bool), typeof(IllustDetailItem));
        public static readonly BindableProperty DownloadingProperty = BindableProperty.Create(
            nameof(Downloading), typeof(bool), typeof(IllustDetailItem));

        public ImageSource Image
        {
            get => (ImageSource)GetValue(ImageProperty);
            set => SetValue(ImageProperty, value);
        }
        public bool Loading
        {
            get => (bool)GetValue(LoadingProperty);
            set => SetValue(LoadingProperty, value);
        }
        public bool Downloading
        {
            get => (bool)GetValue(DownloadingProperty);
            set => SetValue(DownloadingProperty, value);
        }
        public string Id { get; set; }
        public string PreviewUrl { get; set; }
        public string OriginalUrl { get; set; }
    }
}
