﻿using System;
using System.IO;
using System.Net;
using System.Net.Http;
using System.Net.Http.Headers;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using Newtonsoft.Json;
using Pixiview.Illust;
using Xamarin.Forms;

namespace Pixiview.Utils
{
    public class HttpUtility
    {
        public static T LoadObject<T>(string file, string url, string referer,
            bool force = false,
            Action<HttpRequestHeaders> header = null,
            Func<T, string> namehandler = null,
            Func<string, string> action = null)
        {
            string content = null;
            if (!force && file != null && File.Exists(file))
            {
                try
                {
                    content = File.ReadAllText(file);
                }
                catch (Exception ex)
                {
                    App.DebugError("load", $"failed to read file: {file}, error: {ex.Message}");
                }
            }
            if (content == null)
            {
                var response = Download(url, headers =>
                {
                    if (referer != null)
                    {
                        headers.Referrer = new Uri(referer);
                    }
                    headers.Add("User-Agent", Configs.UserAgent);
                    headers.Add("Accept", Configs.AcceptJson);
                    headers.Add("Cookie", Configs.Cookie);
                    if (header == null)
                    {
                        headers.Add("X-User-Id", Configs.UserId);
                    }
                    else
                    {
                        header(headers);
                    }
                });
                if (response == null)
                {
                    return default;
                }
                using (response)
                {
                    try
                    {
                        content = response.Content.ReadAsStringAsync().Result;
                        if (action != null)
                        {
                            content = action(content);
                        }
                    }
                    catch (Exception ex)
                    {
                        App.DebugError("load.strea", $"failed to read stream, error: {ex.Message}");
                        return default;
                    }

                    bool rtn = false;
                    T result = default;
                    if (namehandler != null)
                    {
                        try
                        {
                            result = JsonConvert.DeserializeObject<T>(content);
                            file = namehandler(result);
                            rtn = true;
                        }
                        catch (Exception ex)
                        {
                            App.DebugError("load", $"failed to parse illust JSON object, error: {ex.Message}");
                        }
                    }

                    if (file != null)
                    {
                        try
                        {
                            var folder = Path.GetDirectoryName(file);
                            if (!Directory.Exists(folder))
                            {
                                Directory.CreateDirectory(folder);
                            }
                            File.WriteAllText(file, content, Encoding.UTF8);
                        }
                        catch (Exception ex)
                        {
                            App.DebugError("save", $"failed to save illust JSON object, error: {ex.Message}");
                        }
                    }

                    if (rtn)
                    {
                        return result;
                    }
                }
            }
            try
            {
                return JsonConvert.DeserializeObject<T>(content);
            }
            catch (Exception ex)
            {
                App.DebugError("load", $"failed to parse illust JSON object, error: {ex.Message}");
                return default;
            }
        }

        public static string DownloadImage(string url, string working, string folder)
        {
            try
            {
                var directory = Path.Combine(working, folder);
                if (!Directory.Exists(directory))
                {
                    Directory.CreateDirectory(directory);
                }
                var file = Path.Combine(directory, Path.GetFileName(url));
                App.DebugPrint($"download, url: {url}");
                var response = Download(url, headers =>
                {
                    headers.Referrer = new Uri(Configs.Referer);
                    headers.Add("User-Agent", Configs.UserAgent);
                    headers.Add("Accept", Configs.AcceptImage);
                });
                if (response == null)
                {
                    return null;
                }
                using (response)
                using (var fs = File.OpenWrite(file))
                {
                    response.Content.CopyToAsync(fs).Wait();
                    //if (response.Headers.Date != null)
                    //{
                    //    File.SetLastWriteTimeUtc(file, response.Headers.Date.Value.UtcDateTime);
                    //}
                }
                return file;
            }
            catch (Exception ex)
            {
                App.DebugError("image.download", ex.Message);
                return null;
            }
        }

        private static HttpResponseMessage Download(string url, Action<HttpRequestHeaders> headerAction)
        {
            App.DebugPrint($"GET: {url}");
            var uri = new Uri(url);
            var proxy = Configs.Proxy;
            var handler = new HttpClientHandler
            {
                AutomaticDecompression = DecompressionMethods.GZip | DecompressionMethods.Deflate,
                UseCookies = false
            };
            if (proxy != null)
            {
                handler.Proxy = proxy;
                handler.UseProxy = true;
            }
            var client = new HttpClient(handler)
            {
                BaseAddress = new Uri($"{uri.Scheme}://{uri.Host}"),
                Timeout = TimeSpan.FromSeconds(30)
            };
            return TryCount(() =>
            {
                using (var request = new HttpRequestMessage(HttpMethod.Get, uri.PathAndQuery)
                {
                    Version = new Version(2, 0)
                })
                {
                    var headers = request.Headers;
                    headerAction(headers);
                    if (proxy == null)
                    {
                        headers.Add("x-reverse", "yes");
                    }
                    headers.Add("Accept-Language", Configs.AcceptLanguage);
                    //headers.Add("Accept-Encoding", Configs.AcceptEncoding);
                    return client.SendAsync(request, HttpCompletionOption.ResponseHeadersRead).Result;
                }
            });
        }

        private static T TryCount<T>(Func<T> func, int tryCount = 2)
        {
            int tries = 0;
            while (tries < tryCount)
            {
                try
                {
                    return func();
                }
                catch (Exception ex)
                {
                    tries++;
                    Thread.Sleep(1000);
                    App.DebugError("try.do", $"tries: {tries}, error: {ex.Message}");
                }
            }
            return default;
        }

        public static (long Size, DateTimeOffset LastModified, HttpClient Client) GetUgoiraHeader(string url, string id)
        {
            var uri = new Uri(url);
            var proxy = Configs.Proxy;
            var handler = new HttpClientHandler
            {
                UseCookies = false
            };
            if (proxy != null)
            {
                handler.Proxy = proxy;
                handler.UseProxy = true;
            }
            var client = new HttpClient(handler)
            {
                BaseAddress = new Uri($"{uri.Scheme}://{uri.Host}"),
                Timeout = TimeSpan.FromSeconds(30)
            };
            var response = TryCount(() =>
            {
                using (var request = new HttpRequestMessage(HttpMethod.Head, uri.PathAndQuery)
                {
                    Version = new Version(2, 0)
                })
                {
                    var headers = request.Headers;
                    UgoiraHeaderAction(headers, id);
                    headers.Add("Accept-Encoding", "gzip, deflate");
                    headers.Add("Accept-Language", Configs.AcceptLanguage);
                    return client.SendAsync(request, HttpCompletionOption.ResponseHeadersRead).Result;
                }
            });

            var size = response.Content.Headers.ContentLength.Value;
            var lastModified = response.Content.Headers.LastModified.Value;

            return (size, lastModified, client);
        }

        public static long DownloadUgoiraImage(HttpClient client, string url, string id, DateTimeOffset lastModified, long from, long to, Stream stream)
        {
            var uri = new Uri(url);
            var response = TryCount(() =>
            {
                using (var request = new HttpRequestMessage(HttpMethod.Get, uri.PathAndQuery)
                {
                    Version = new Version(2, 0)
                })
                {
                    var headers = request.Headers;
                    UgoiraHeaderAction(headers, id);
                    headers.Add("Accept-Encoding", "identity");
                    headers.IfRange = new RangeConditionHeaderValue(lastModified);
                    headers.Range = new RangeHeaderValue(from, to);
                    headers.Add("Accept-Language", Configs.AcceptLanguage);
                    return client.SendAsync(request, HttpCompletionOption.ResponseHeadersRead).Result;
                }
            });

            var length = response.Content.Headers.ContentLength.Value;
            response.Content.CopyToAsync(stream).Wait();
            return length;
        }

        private static void UgoiraHeaderAction(HttpRequestHeaders headers, string id)
        {
            headers.Add("Accept", "*/*");
            headers.Add("Origin", Configs.Referer);
            headers.Referrer = new Uri(string.Format(Configs.RefererIllust, id));
            headers.Add("User-Agent", Configs.UserAgent);
        }
    }

    public class Ugoira
    {
        private readonly IllustUgoiraBody ugoira;
        private readonly IllustDetailItem detailItem;
        private readonly ImageSource[] frames;
        private readonly Timer timer;
        private int index = 0;

        public bool IsPlaying { get; private set; }
        public readonly int FrameCount;
        public event EventHandler<UgoiraEventArgs> FrameChanged;

        public Ugoira(IllustUgoiraData illust, IllustDetailItem item)
        {
            ugoira = illust.body;
            detailItem = item;
            frames = new ImageSource[ugoira.frames.Length];
            FrameCount = frames.Length;
            timer = new Timer(OnTimerCallback, null, Timeout.Infinite, Timeout.Infinite);

            Task.Run(LoadFrames);
        }

        public void TogglePlay(bool flag)
        {
            if (IsPlaying == flag)
            {
                return;
            }
            if (flag)
            {
                IsPlaying = true;
                timer.Change(0, Timeout.Infinite);
            }
            else
            {
                IsPlaying = false;
                timer.Change(Timeout.Infinite, Timeout.Infinite);
            }
        }

        public void ToggleFrame(int frame)
        {
            if (IsPlaying)
            {
                // TODO: doesn't support change current frame when playing
                return;
            }
            if (frame < 0 || frame >= frames.Length)
            {
                return;
            }
            var image = frames[frame];
            if (image != null)
            {
                index = frame;
                FrameChanged?.Invoke(this, new UgoiraEventArgs
                {
                    DetailItem = detailItem,
                    Image = image,
                    FrameIndex = frame
                });
            }
        }

        private void OnTimerCallback(object state)
        {
            if (!IsPlaying)
            {
                return;
            }

            ImageSource frame;
            var i = index;
            var info = ugoira.frames[i];
            while ((frame = frames[i]) == null)
            {
                // not downloaded yet, waiting...
                Thread.Sleep(100);
            }
            FrameChanged?.Invoke(this, new UgoiraEventArgs
            {
                DetailItem = detailItem,
                Image = frame,
                FrameIndex = i
            });
            i++;
            if (i >= frames.Length)
            {
                i = 0;
            }
            index = i;
            timer.Change(info.delay, Timeout.Infinite);
        }

        private const int BUFFER_SIZE = 300000;

        private void LoadFrames()
        {
            var zip = Path.GetFileName(ugoira.src);
            bool download = false;
            for (var i = 0; i < ugoira.frames.Length; i++)
            {
                var frame = ugoira.frames[i];
                var image = Stores.LoadUgoiraImage(zip, frame.file);
                if (image != null)
                {
                    frames[i] = image;
                }
                else
                {
                    download = true;
                    break;
                }
            }

            if (download)
            {
                // need download
                var url = ugoira.src;
                var id = detailItem.Id;
                var (size, lastModified, client) = HttpUtility.GetUgoiraHeader(url, id);
                App.DebugPrint($"starting download ugoira: {size} bytes, last modified: {lastModified}");

                var data = new byte[size];
                using (var ms = new MemoryStream(data))
                {
                    var index = 0;
                    for (var i = 0; ; i += BUFFER_SIZE)
                    {
                        long to;
                        if (i + BUFFER_SIZE > size)
                        {
                            to = size - 1;
                        }
                        else
                        {
                            to = i + BUFFER_SIZE - 1;
                        }
                        HttpUtility.DownloadUgoiraImage(client, url, id, lastModified, i, to, ms);
                        var last = ms.Position;
                        while (true)
                        {
                            var pos = ExtractImage(zip, data, index, last);
                            if (pos > index)
                            {
                                index = pos;
                            }
                            else
                            {
                                break;
                            }
                        }

                        if (i + BUFFER_SIZE > size)
                        {
                            break;
                        }
                    }
                }
            }
        }

        private int ExtractImage(string zip, byte[] data, int index, long last)
        {
            var i = index;
            if (i + 30 > last)
            {
                return index;
            }
            if (data[i] != 0x50 || data[i + 1] != 0x4b ||
                data[i + 2] != 0x03 || data[i + 3] != 0x04)
            {
                App.DebugPrint($"extract complete, header: {BitConverter.ToInt32(data, i):x8}");
                return index;
            }
            i += 8; // signature(4) & version(2) & flags(2)
            if (data[i] != 0 || data[i + 1] != 0)
            {
                App.DebugError("extract.image", $"doesn't support compressed data: {BitConverter.ToInt16(data, i):x4}");
                return index;
            }
            i += 10; // compression(2) & mod-time(2) & mod-date(2) & crc-32(4)
            int size = BitConverter.ToInt32(data, i);
            i += 4; // size(4)
            int rawSize = BitConverter.ToInt32(data, i);
            if (size != rawSize)
            {
                App.DebugError("extract.image", $"data seems to be compressed: {size} ({rawSize}) bytes");
                return index;
            }
            i += 4; // rawSize(4)
            int filenameLength = BitConverter.ToInt16(data, i);
            i += 2; // filename length(2)
            int extraLength = BitConverter.ToInt16(data, i);
            i += 2; // extra length(2)

            if (i + filenameLength + extraLength + size > last)
            {
                App.DebugPrint($"download is not completed, index: {index}, size: {size}, last: {last}");
                return index;
            }

            var filename = Encoding.UTF8.GetString(data, i, filenameLength);
            i += filenameLength + extraLength;  // filename & extra

            // content
            var content = new byte[size];
            Array.Copy(data, i, content, 0, size);
            i += size;

            var file = Stores.SaveUgoiraImage(zip, filename, content);
            if (file != null)
            {
                for (var n = 0; n < ugoira.frames.Length; n++)
                {
                    if (ugoira.frames[n].file == filename)
                    {
                        App.DebugPrint($"load frame: {filename}");
                        frames[n] = ImageSource.FromFile(file);
                        break;
                    }
                }
            }
            return i;
        }
    }

    public class UgoiraEventArgs : EventArgs
    {
        public IllustDetailItem DetailItem { get; set; }
        public ImageSource Image { get; set; }
        public int FrameIndex { get; set; }
    }
}
